#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

import elasticapm

redis = pytest.importorskip("redis.asyncio")  # isort:skip

import os

import pytest_asyncio

from elasticapm.conf.constants import TRANSACTION
from elasticapm.traces import capture_span

pytestmark = [pytest.mark.asyncio, pytest.mark.redis]

if "REDIS_HOST" not in os.environ:
    pytestmark.append(pytest.mark.skip("Skipping redis tests, no REDIS_HOST environment variable set"))


@pytest_asyncio.fixture()
async def redis_async_conn():
    _host = os.environ["REDIS_HOST"]
    _port = os.environ.get("REDIS_PORT", 6379)
    conn = await redis.Redis.from_url(f"redis://{_host}:{_port}")

    yield conn

    await conn.close(close_connection_pool=True)


@pytest.mark.integrationtest
async def test_ping(instrument, elasticapm_client, redis_async_conn):
    # The PING command is sent as a byte string, so this tests if we can handle
    # the command both as a str and as a bytes. See #1307
    elasticapm_client.begin_transaction("transaction.test")
    await redis_async_conn.ping()
    elasticapm_client.end_transaction("test")
    transaction = elasticapm_client.events[TRANSACTION][0]
    span = elasticapm_client.spans_for_transaction(transaction)[0]
    assert span["name"] == "PING"
    assert span["duration"] > 0.2  # sanity test to ensure we measure the actual call


@pytest.mark.integrationtest
async def test_pipeline(instrument, elasticapm_client, redis_async_conn):
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_pipeline", "test"):
        pipeline = redis_async_conn.pipeline()
        pipeline.rpush("mykey", "a", "b")
        pipeline.expire("mykey", 1000)
        await pipeline.execute()
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    assert spans[0]["name"] == "Pipeline.execute"
    assert spans[0]["type"] == "db"
    assert spans[0]["subtype"] == "redis"
    assert spans[0]["action"] == "query"
    assert spans[0]["context"]["destination"] == {
        "address": os.environ.get("REDIS_HOST", "localhost"),
        "port": int(os.environ.get("REDIS_PORT", 6379)),
        "service": {"name": "", "resource": "redis", "type": ""},
    }
    assert spans[0]["duration"] > 0.2  # sanity test to ensure we measure the actual call

    assert spans[1]["name"] == "test_pipeline"
    assert spans[1]["type"] == "test"

    assert len(spans) == 2


@pytest.mark.integrationtest
async def test_redis_client(instrument, elasticapm_client, redis_async_conn):
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_redis_client", "test"):
        await redis_async_conn.rpush("mykey", "a", "b")
        await redis_async_conn.expire("mykey", 1000)
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    spans = sorted(spans, key=lambda x: x["name"])

    assert {t["name"] for t in spans} == {"test_redis_client", "RPUSH", "EXPIRE"}

    assert spans[0]["name"] == "EXPIRE"
    assert spans[0]["type"] == "db"
    assert spans[0]["subtype"] == "redis"
    assert spans[0]["action"] == "query"
    assert spans[0]["context"]["destination"] == {
        "address": os.environ.get("REDIS_HOST", "localhost"),
        "port": int(os.environ.get("REDIS_PORT", 6379)),
        "service": {"name": "", "resource": "redis", "type": ""},
    }
    assert spans[0]["duration"] > 0.2  # sanity test to ensure we measure the actual call

    assert spans[1]["name"] == "RPUSH"
    assert spans[1]["type"] == "db"
    assert spans[1]["subtype"] == "redis"
    assert spans[1]["action"] == "query"
    assert spans[1]["context"]["destination"] == {
        "address": os.environ.get("REDIS_HOST", "localhost"),
        "port": int(os.environ.get("REDIS_PORT", 6379)),
        "service": {"name": "", "resource": "redis", "type": ""},
    }
    assert spans[1]["duration"] > 0.2  # sanity test to ensure we measure the actual call

    assert spans[2]["name"] == "test_redis_client"
    assert spans[2]["type"] == "test"

    assert len(spans) == 3


@pytest.mark.integrationtest
async def test_publish_subscribe_async(instrument, elasticapm_client, redis_async_conn):
    elasticapm_client.begin_transaction("transaction.test")
    pubsub = redis_async_conn.pubsub()
    with capture_span("test_publish_subscribe", "test"):
        # publish
        await redis_async_conn.publish("mykey", "a")

        # subscribe
        await pubsub.subscribe("mykey")

    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    expected_signatures = {"test_publish_subscribe", "PUBLISH", "SUBSCRIBE"}

    assert {t["name"] for t in spans} == expected_signatures

    assert spans[0]["name"] == "PUBLISH"
    assert spans[0]["type"] == "db"
    assert spans[0]["subtype"] == "redis"
    assert spans[0]["action"] == "query"
    assert spans[0]["context"]["destination"] == {
        "address": os.environ.get("REDIS_HOST", "localhost"),
        "port": int(os.environ.get("REDIS_PORT", 6379)),
        "service": {"name": "", "resource": "redis", "type": ""},
    }
    assert spans[0]["duration"] > 0.2  # sanity test to ensure we measure the actual call

    assert spans[1]["name"] == "SUBSCRIBE"
    assert spans[1]["type"] == "db"
    assert spans[1]["subtype"] == "redis"
    assert spans[1]["action"] == "query"
    assert spans[1]["context"]["destination"] == {
        "address": os.environ.get("REDIS_HOST", "localhost"),
        "port": int(os.environ.get("REDIS_PORT", 6379)),
        "service": {"name": "", "resource": "redis", "type": ""},
    }
    assert spans[1]["duration"] > 0.2  # sanity test to ensure we measure the actual call

    assert spans[2]["name"] == "test_publish_subscribe"
    assert spans[2]["type"] == "test"

    assert len(spans) == 3


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        pytest.param({"transaction_max_spans": 1}),
    ],
    indirect=True,
)
@pytest.mark.integrationtest
async def test_dropped(instrument, elasticapm_client, redis_async_conn):
    # Test that our instrumentation doesn't blow up for dropped spans
    elasticapm_client.begin_transaction("transaction.test")
    async with elasticapm.async_capture_span("bla"):
        pass
    await redis_async_conn.ping()
    elasticapm_client.end_transaction("test")
    transaction = elasticapm_client.events[TRANSACTION][0]
    span = elasticapm_client.spans_for_transaction(transaction)[0]
    assert span["name"] == "bla"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import mock
import urllib3

import elasticapm.utils.cloud

GCP_DATA = b"""{
    "instance": {
        "id": 4306570268266786072,
        "machineType": "projects/513326162531/machineTypes/n1-standard-1",
        "name": "basepi-test",
        "zone": "projects/513326162531/zones/us-west3-a"
    },
    "project": {"numericProjectId": 513326162531, "projectId": "elastic-apm"}
}"""

AZURE_DATA = b"""{
    "location": "westus2",
    "name": "basepi-test",
    "resourceGroupName": "basepi-testing",
    "subscriptionId": "7657426d-c4c3-44ac-88a2-3b2cd59e6dba",
    "vmId": "e11ebedc-019d-427f-84dd-56cd4388d3a8",
    "vmScaleSetName": "",
    "vmSize": "Standard_D2s_v3",
    "zone": ""
}"""

AWS_DATA = b"""{
    "accountId": "946960629917",
    "architecture": "x86_64",
    "availabilityZone": "us-east-2a",
    "billingProducts": null,
    "devpayProductCodes": null,
    "marketplaceProductCodes": null,
    "imageId": "ami-07c1207a9d40bc3bd",
    "instanceId": "i-0ae894a7c1c4f2a75",
    "instanceType": "t2.medium",
    "kernelId": null,
    "pendingTime": "2020-06-12T17:46:09Z",
    "privateIp": "172.31.0.212",
    "ramdiskId": null,
    "region": "us-east-2",
    "version": "2017-09-30"
}"""


@mock.patch("socket.create_connection")
def test_aws_metadata(mock_socket, monkeypatch):
    class MockPoolManager:
        def request(self, *args, **kwargs):
            self.data = AWS_DATA
            return self

    monkeypatch.setattr(urllib3, "PoolManager", MockPoolManager)

    metadata = elasticapm.utils.cloud.aws_metadata()
    assert metadata == {
        "account": {"id": "946960629917"},
        "instance": {"id": "i-0ae894a7c1c4f2a75"},
        "availability_zone": "us-east-2a",
        "machine": {"type": "t2.medium"},
        "provider": "aws",
        "region": "us-east-2",
    }


@mock.patch("socket.getaddrinfo")
def test_gcp_metadata(mock_socket, monkeypatch):
    class MockPoolManager:
        def request(self, *args, **kwargs):
            self.data = GCP_DATA
            return self

    monkeypatch.setattr(urllib3, "PoolManager", MockPoolManager)

    metadata = elasticapm.utils.cloud.gcp_metadata()
    assert metadata == {
        "provider": "gcp",
        "instance": {"id": "4306570268266786072", "name": "basepi-test"},
        "project": {"id": "elastic-apm"},
        "availability_zone": "us-west3-a",
        "region": "us-west3",
        "machine": {"type": "n1-standard-1"},
    }


@mock.patch("socket.create_connection")
def test_azure_metadata(mock_socket, monkeypatch):
    class MockPoolManager:
        def request(self, *args, **kwargs):
            self.data = AZURE_DATA
            return self

    monkeypatch.setattr(urllib3, "PoolManager", MockPoolManager)

    metadata = elasticapm.utils.cloud.azure_metadata()
    assert metadata == {
        "account": {"id": "7657426d-c4c3-44ac-88a2-3b2cd59e6dba"},
        "instance": {"id": "e11ebedc-019d-427f-84dd-56cd4388d3a8", "name": "basepi-test"},
        "project": {"name": "basepi-testing"},
        "machine": {"type": "Standard_D2s_v3"},
        "provider": "azure",
        "region": "westus2",
    }


def test_azure_app_service_metadata():
    env_template = {
        "WEBSITE_RESOURCE_GROUP": "resource_group",
        "WEBSITE_SITE_NAME": "site_name",
        "WEBSITE_INSTANCE_ID": "instance_id",
    }
    for owner_name in (
        "f5940f10-2e30-3e4d-a259-63451ba6dae4+elastic-apm-AustraliaEastwebspace",
        "f5940f10-2e30-3e4d-a259-63451ba6dae4+appsvc_linux_australiaeast-AustraliaEastwebspace-Linux",
    ):
        env = env_template.copy()
        env["WEBSITE_OWNER_NAME"] = owner_name
        with mock.patch.dict(os.environ, env):
            metadata = elasticapm.utils.cloud.azure_app_service_metadata()
            assert metadata == {
                "account": {"id": "f5940f10-2e30-3e4d-a259-63451ba6dae4"},
                "instance": {"id": "instance_id", "name": "site_name"},
                "project": {"name": "resource_group"},
                "provider": "azure",
                "region": "AustraliaEast",
            }


def test_azure_app_service_metadata_wrong_format():
    env = {
        # the - before "elastic" is faulty
        "WEBSITE_OWNER_NAME": "f5940f10-2e30-3e4d-a259-63451ba6dae4-elastic-apm-AustraliaEastwebspace",
        "WEBSITE_RESOURCE_GROUP": "resource_group",
        "WEBSITE_SITE_NAME": "site_name",
        "WEBSITE_INSTANCE_ID": "instance_id",
    }
    with mock.patch.dict(os.environ, env):
        metadata = elasticapm.utils.cloud.azure_app_service_metadata()
        assert metadata == {}


def test_azure_app_service_metadata_empty():
    with mock.patch.dict(os.environ, {}):
        metadata = elasticapm.utils.cloud.azure_app_service_metadata()
        assert metadata == {}


def test_azure_app_service_metadata_not_complete():
    env_template = {
        "WEBSITE_OWNER_NAME": "f5940f10-2e30-3e4d-a259-63451ba6dae4+elastic-apm-AustraliaEastwebspace",
        "WEBSITE_RESOURCE_GROUP": "resource_group",
        "WEBSITE_SITE_NAME": "site_name",
        "WEBSITE_INSTANCE_ID": "instance_id",
    }
    for k in env_template.keys():
        env = env_template.copy()
        with mock.patch.dict(os.environ, env):
            os.environ[k] = ""
            metadata = elasticapm.utils.cloud.azure_app_service_metadata()
            assert metadata == {}
            del os.environ[k]
            metadata = elasticapm.utils.cloud.azure_app_service_metadata()
            assert metadata == {}



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sqlite3

from elasticapm.conf.constants import TRANSACTION


def test_connect(instrument, elasticapm_client):
    elasticapm_client.begin_transaction("transaction.test")

    sqlite3.connect(":memory:")
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    assert spans[0]["name"] == "sqlite3.connect :memory:"
    assert spans[0]["type"] == "db"
    assert spans[0]["subtype"] == "sqlite"
    assert spans[0]["action"] == "connect"


def test_cursor(instrument, elasticapm_client):
    conn = sqlite3.connect(":memory:")

    elasticapm_client.begin_transaction("transaction.test")
    cursor = conn.cursor()
    cursor.execute("CREATE TABLE testdb (id integer, username text)")
    cursor.execute('INSERT INTO testdb VALUES (1, "Ron")')
    cursor.executemany("INSERT INTO testdb VALUES (?, ?)", ((2, "Rasmus"), (3, "Shay")))
    cursor.execute("DROP TABLE testdb")
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    expected_signatures = {"CREATE TABLE", "INSERT INTO testdb", "DROP TABLE"}

    assert {t["name"] for t in spans} == expected_signatures

    assert spans[0]["name"] == "CREATE TABLE"
    assert spans[0]["type"] == "db"
    assert spans[0]["subtype"] == "sqlite"
    assert spans[0]["action"] == "query"

    assert spans[1]["name"] == "INSERT INTO testdb"
    assert spans[1]["type"] == "db"
    assert spans[1]["subtype"] == "sqlite"
    assert spans[1]["action"] == "query"

    assert spans[2]["name"] == "INSERT INTO testdb"
    assert spans[2]["type"] == "db"
    assert spans[2]["subtype"] == "sqlite"
    assert spans[2]["action"] == "query"

    assert spans[3]["name"] == "DROP TABLE"
    assert spans[3]["type"] == "db"
    assert spans[3]["subtype"] == "sqlite"
    assert spans[3]["action"] == "query"

    assert len(spans) == 4


def test_nonstandard_connection_execute(instrument, elasticapm_client):
    conn = sqlite3.connect(":memory:")

    elasticapm_client.begin_transaction("transaction.test")
    conn.execute("CREATE TABLE testdb (id integer, username text)")
    conn.execute('INSERT INTO testdb VALUES (1, "Ron")')
    conn.executemany("INSERT INTO testdb VALUES (?, ?)", ((2, "Rasmus"), (3, "Shay")))
    conn.execute("DROP TABLE testdb")
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    expected_signatures = {"CREATE TABLE", "INSERT INTO testdb", "DROP TABLE"}

    assert {t["name"] for t in spans} == expected_signatures

    assert spans[0]["name"] == "CREATE TABLE"
    assert spans[0]["type"] == "db"
    assert spans[0]["subtype"] == "sqlite"
    assert spans[0]["action"] == "query"

    assert spans[1]["name"] == "INSERT INTO testdb"
    assert spans[1]["type"] == "db"
    assert spans[1]["subtype"] == "sqlite"
    assert spans[1]["action"] == "query"

    assert spans[2]["name"] == "INSERT INTO testdb"
    assert spans[2]["type"] == "db"
    assert spans[2]["subtype"] == "sqlite"
    assert spans[2]["action"] == "query"

    assert spans[3]["name"] == "DROP TABLE"
    assert spans[3]["type"] == "db"
    assert spans[3]["subtype"] == "sqlite"
    assert spans[3]["action"] == "query"

    assert len(spans) == 4


def test_truncate_long_sql(instrument, elasticapm_client):
    conn = sqlite3.connect(":memory:")

    conn.execute("CREATE TABLE testdb (id integer, username text)")
    conn.execute('INSERT INTO testdb VALUES (1, "Ron")')
    elasticapm_client.begin_transaction("transaction.test")
    conn.execute("""SELECT id, username FROM testdb WHERE username = "%s";""" % ("x" * 10010))
    elasticapm_client.end_transaction("MyView")
    conn.execute("DROP TABLE testdb")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    assert len(spans[0]["context"]["db"]["statement"]) == 10000
    assert spans[0]["context"]["db"]["statement"].endswith("...")



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import logging.handlers
import sys
import warnings
from logging import LogRecord

import ecs_logging
import pytest
import structlog

from elasticapm.conf import Config
from elasticapm.conf.constants import ERROR
from elasticapm.handlers.logging import Formatter, LoggingFilter, LoggingHandler
from elasticapm.handlers.structlog import structlog_processor
from elasticapm.traces import capture_span
from elasticapm.utils.stacks import iter_stack_frames
from tests.fixtures import TempStoreClient


@pytest.fixture()
def logger(elasticapm_client):
    elasticapm_client.config.include_paths = ["tests", "elasticapm"]
    handler = LoggingHandler(elasticapm_client)
    logger = logging.getLogger(__name__)
    logger.handlers = []
    logger.addHandler(handler)
    logger.client = elasticapm_client
    logger.level = logging.INFO
    return logger


def test_logger_basic(logger):
    logger.error("This is a test error")

    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]
    assert event["log"]["logger_name"] == __name__
    assert event["log"]["level"] == "error"
    assert event["log"]["message"] == "This is a test error"
    assert "stacktrace" in event["log"]
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test error"


def test_logger_warning(logger):
    logger.warning("This is a test warning")
    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]
    assert event["log"]["logger_name"] == __name__
    assert event["log"]["level"] == "warning"
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test warning"


def test_logger_extra_data(logger):
    logger.info("This is a test info with a url", extra=dict(data=dict(url="http://example.com")))
    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]
    assert event["context"]["custom"]["url"] == "http://example.com"
    assert "stacktrace" in event["log"]
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test info with a url"


def test_logger_exc_info(logger):
    try:
        raise ValueError("This is a test ValueError")
    except ValueError:
        logger.info("This is a test info with an exception", exc_info=True)

    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]

    # assert event['message'] == 'This is a test info with an exception'
    assert "exception" in event
    assert "stacktrace" in event["exception"]
    exc = event["exception"]
    assert exc["type"] == "ValueError"
    assert exc["message"] == "ValueError: This is a test ValueError"
    assert "param_message" in event["log"]
    assert event["log"]["message"] == "This is a test info with an exception"


def test_message_params(logger):
    logger.info("This is a test of %s", "args")
    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["message"] == "This is a test of args"
    assert event["log"]["param_message"] == "This is a test of %s"


def test_record_stack(logger):
    logger.info("This is a test of stacks", extra={"stack": True})
    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]
    frames = event["log"]["stacktrace"]
    assert len(frames) != 1
    frame = frames[0]
    assert frame["module"] == __name__
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test of stacks"
    assert event["culprit"] == "tests.handlers.logging.logging_tests.test_record_stack"
    assert event["log"]["message"] == "This is a test of stacks"


def test_no_record_stack(logger):
    logger.info("This is a test of no stacks", extra={"stack": False})
    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]
    assert event.get("culprit") == None
    assert event["log"]["message"] == "This is a test of no stacks"
    assert "stacktrace" not in event["log"]
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test of no stacks"


def test_no_record_stack_via_config(logger):
    logger.client.config.auto_log_stacks = False
    logger.info("This is a test of no stacks")
    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]
    assert event.get("culprit") == None
    assert event["log"]["message"] == "This is a test of no stacks"
    assert "stacktrace" not in event["log"]
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test of no stacks"


def test_explicit_stack(logger):
    logger.info("This is a test of stacks", extra={"stack": iter_stack_frames()})
    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]
    assert "culprit" in event, event
    assert event["culprit"] == "tests.handlers.logging.logging_tests.test_explicit_stack"
    assert "message" in event["log"], event
    assert event["log"]["message"] == "This is a test of stacks"
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test of stacks"
    assert "stacktrace" in event["log"]


def test_extra_culprit(logger):
    logger.info("This is a test of stacks", extra={"culprit": "foo.bar"})
    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]
    assert event["culprit"] == "foo.bar"
    assert "culprit" not in event["context"]["custom"]


def test_logger_exception(logger):
    try:
        raise ValueError("This is a test ValueError")
    except ValueError:
        logger.exception("This is a test with an exception", extra={"stack": True})

    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]

    assert event["log"]["message"] == "This is a test with an exception"
    assert "stacktrace" in event["log"]
    assert "exception" in event
    exc = event["exception"]
    assert exc["type"] == "ValueError"
    assert exc["message"] == "ValueError: This is a test ValueError"
    assert "param_message" in event["log"]
    assert event["log"]["message"] == "This is a test with an exception"


def test_client_arg(elasticapm_client):
    handler = LoggingHandler(elasticapm_client)
    assert handler.client == elasticapm_client


def test_client_kwarg(elasticapm_client):
    handler = LoggingHandler(client=elasticapm_client)
    assert handler.client == elasticapm_client


def test_logger_setup():
    handler = LoggingHandler(
        server_url="foo", service_name="bar", secret_token="baz", metrics_interval="0ms", client_cls=TempStoreClient
    )
    client = handler.client
    assert client.config.server_url == "foo"
    assert client.config.service_name == "bar"
    assert client.config.secret_token == "baz"
    assert handler.level == logging.NOTSET


def test_logging_handler_emit_error(capsys, elasticapm_client):
    handler = LoggingHandler(elasticapm_client)
    handler._emit = lambda: 1 / 0
    handler.emit(LogRecord("x", 1, "/ab/c/", 10, "Oops", (), None))
    out, err = capsys.readouterr()
    assert "Top level ElasticAPM exception caught" in err
    assert "Oops" in err


def test_logging_handler_dont_emit_elasticapm(capsys, elasticapm_client):
    handler = LoggingHandler(elasticapm_client)
    handler.emit(LogRecord("elasticapm.errors", 1, "/ab/c/", 10, "Oops", (), None))
    out, err = capsys.readouterr()
    assert "Oops" in err


def test_logging_handler_emit_error_non_str_message(capsys, elasticapm_client):
    handler = LoggingHandler(elasticapm_client)
    handler._emit = lambda: 1 / 0
    handler.emit(LogRecord("x", 1, "/ab/c/", 10, ValueError("oh no"), (), None))
    out, err = capsys.readouterr()
    assert "Top level ElasticAPM exception caught" in err
    assert "oh no" in err


def test_arbitrary_object(logger):
    logger.error(["a", "list", "of", "strings"])
    assert len(logger.client.events) == 1
    event = logger.client.events[ERROR][0]
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "['a', 'list', 'of', 'strings']"


def test_logging_filter_no_span(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    f = LoggingFilter()
    record = logging.LogRecord(__name__, logging.DEBUG, __file__, 252, "dummy_msg", [], None)
    f.filter(record)
    assert record.elasticapm_transaction_id == transaction.id
    assert record.elasticapm_service_name == transaction.tracer.config.service_name
    assert record.elasticapm_service_environment == transaction.tracer.config.environment
    assert record.elasticapm_trace_id == transaction.trace_parent.trace_id
    assert record.elasticapm_span_id is None
    assert record.elasticapm_labels


def test_structlog_processor_no_span(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    event_dict = {}
    new_dict = structlog_processor(None, None, event_dict)
    assert new_dict["transaction.id"] == transaction.id
    assert new_dict["trace.id"] == transaction.trace_parent.trace_id
    assert "span.id" not in new_dict


@pytest.mark.parametrize("elasticapm_client", [{"transaction_max_spans": 5}], indirect=True)
def test_logging_filter_span(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    with capture_span("test") as span:
        f = LoggingFilter()
        record = logging.LogRecord(__name__, logging.DEBUG, __file__, 252, "dummy_msg", [], None)
        f.filter(record)
        assert record.elasticapm_transaction_id == transaction.id
        assert record.elasticapm_service_name == transaction.tracer.config.service_name
        assert record.elasticapm_service_environment == transaction.tracer.config.environment
        assert record.elasticapm_trace_id == transaction.trace_parent.trace_id
        assert record.elasticapm_span_id == span.id
        assert record.elasticapm_labels

    # Capture too many spans so we start dropping
    for i in range(10):
        with capture_span("drop"):
            pass

    # Test logging with DroppedSpan
    with capture_span("drop") as span:
        record = logging.LogRecord(__name__, logging.DEBUG, __file__, 252, "dummy_msg2", [], None)
        f.filter(record)
        assert record.elasticapm_transaction_id == transaction.id
        assert record.elasticapm_service_name == transaction.tracer.config.service_name
        assert record.elasticapm_service_environment == transaction.tracer.config.environment
        assert record.elasticapm_trace_id == transaction.trace_parent.trace_id
        assert record.elasticapm_span_id is None
        assert record.elasticapm_labels


@pytest.mark.parametrize("elasticapm_client", [{"transaction_max_spans": 5}], indirect=True)
def test_structlog_processor_span(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    with capture_span("test") as span:
        event_dict = {}
        new_dict = structlog_processor(None, None, event_dict)
        assert new_dict["transaction.id"] == transaction.id
        assert new_dict["service.name"] == transaction.tracer.config.service_name
        assert new_dict["service.environment"] == transaction.tracer.config.environment
        assert new_dict["trace.id"] == transaction.trace_parent.trace_id
        assert new_dict["span.id"] == span.id

    # Capture too many spans so we start dropping
    for i in range(10):
        with capture_span("drop"):
            pass

    # Test logging with DroppedSpan
    with capture_span("drop") as span:
        event_dict = {}
        new_dict = structlog_processor(None, None, event_dict)
        assert new_dict["transaction.id"] == transaction.id
        assert new_dict["service.name"] == transaction.tracer.config.service_name
        assert new_dict["service.environment"] == transaction.tracer.config.environment
        assert new_dict["trace.id"] == transaction.trace_parent.trace_id
        assert "span.id" not in new_dict


def test_automatic_log_record_factory_install(elasticapm_client):
    """
    Use the elasticapm_client fixture to load the client, which in turn installs
    the log_record_factory. Check to make sure it happened.
    """
    transaction = elasticapm_client.begin_transaction("test")
    with capture_span("test") as span:
        record_factory = logging.getLogRecordFactory()
        record = record_factory(__name__, logging.DEBUG, __file__, 252, "dummy_msg", [], None)
        assert record.elasticapm_transaction_id == transaction.id
        assert record.elasticapm_service_name == transaction.tracer.config.service_name
        assert record.elasticapm_service_environment == transaction.tracer.config.environment
        assert record.elasticapm_trace_id == transaction.trace_parent.trace_id
        assert record.elasticapm_span_id == span.id
        assert record.elasticapm_labels


def test_formatter():
    record = logging.LogRecord(__name__, logging.DEBUG, __file__, 252, "dummy_msg", [], None)
    formatter = Formatter()
    formatted_record = formatter.format(record)
    assert "| elasticapm" in formatted_record
    assert hasattr(record, "elasticapm_transaction_id")
    assert hasattr(record, "elasticapm_service_name")
    assert hasattr(record, "elasticapm_service_environment")
    record = logging.LogRecord(__name__, logging.DEBUG, __file__, 252, "dummy_msg", [], None)
    formatted_time = formatter.formatTime(record)
    assert formatted_time
    assert hasattr(record, "elasticapm_transaction_id")
    assert hasattr(record, "elasticapm_service_name")
    assert hasattr(record, "elasticapm_service_environment")


def test_logging_handler_no_client(recwarn):
    # In 6.0, this should be changed to expect a ValueError instead of a log
    warnings.simplefilter("always")
    LoggingHandler(transport_class="tests.fixtures.DummyTransport")
    while True:
        # If we never find our desired warning this will eventually throw an
        # AssertionError
        w = recwarn.pop(DeprecationWarning)
        if "LoggingHandler requires a Client instance" in w.message.args[0]:
            return True


@pytest.mark.parametrize(
    "elasticapm_client,expected",
    [
        ({}, logging.NOTSET),
        ({"log_level": "off"}, 1000),
        ({"log_level": "trace"}, 5),
        ({"log_level": "debug"}, logging.DEBUG),
        ({"log_level": "info"}, logging.INFO),
        ({"log_level": "WARNING"}, logging.WARNING),
        ({"log_level": "errOr"}, logging.ERROR),
        ({"log_level": "CRITICAL"}, logging.CRITICAL),
    ],
    indirect=["elasticapm_client"],
)
def test_log_level_config(elasticapm_client, expected):
    logger = logging.getLogger("elasticapm")
    assert logger.level == expected


def test_log_file(elasticapm_client_log_file):
    logger = logging.getLogger("elasticapm")
    found = False
    for handler in logger.handlers:
        if isinstance(handler, logging.handlers.RotatingFileHandler):
            found = True
    assert found


@pytest.mark.parametrize("elasticapm_client_log_file", [{"log_ecs_reformatting": "override"}], indirect=True)
def test_log_ecs_reformatting(elasticapm_client_log_file):
    logger = logging.getLogger()
    assert isinstance(logger.handlers[0].formatter, ecs_logging.StdlibFormatter)
    assert isinstance(structlog.get_config()["processors"][-1], ecs_logging.StructlogFormatter)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

pytest.importorskip("httplib2")  # isort:skip

import urllib.parse

import httplib2

from elasticapm.conf import constants
from elasticapm.conf.constants import TRANSACTION
from elasticapm.traces import capture_span
from elasticapm.utils.disttracing import TraceParent

pytestmark = pytest.mark.httplib2


@pytest.mark.parametrize("args", [(), ("GET",), ("GET", None, {}), ("GET", None, None)])
def test_httplib2_instrumentation(instrument, elasticapm_client, waiting_httpserver, args):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_request", "test"):
        httplib2.Http().request(url, *args)
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert spans[0]["name"].startswith("GET 127.0.0.1:")
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert url == spans[0]["context"]["http"]["url"]
    assert 200 == spans[0]["context"]["http"]["status_code"]
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % parsed_url.port,
        "type": "",
    }
    assert spans[0]["outcome"] == "success"

    assert constants.TRACEPARENT_HEADER_NAME in waiting_httpserver.requests[0].headers
    trace_parent = TraceParent.from_string(
        waiting_httpserver.requests[0].headers[constants.TRACEPARENT_HEADER_NAME],
        tracestate_string=waiting_httpserver.requests[0].headers[constants.TRACESTATE_HEADER_NAME],
    )
    assert trace_parent.trace_id == transactions[0]["trace_id"]
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict


@pytest.mark.parametrize("status_code", [400, 500])
def test_httplib2_instrumentation_error(instrument, elasticapm_client, waiting_httpserver, status_code):
    waiting_httpserver.serve_content("", code=status_code)
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_request", "test"):
        httplib2.Http().request(url, "GET")
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert spans[0]["name"].startswith("GET 127.0.0.1:")
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert url == spans[0]["context"]["http"]["url"]
    assert status_code == spans[0]["context"]["http"]["status_code"]
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % parsed_url.port,
        "type": "",
    }
    assert spans[0]["outcome"] == "failure"


@pytest.mark.parametrize(
    "is_sampled", [pytest.param(True, id="is_sampled-True"), pytest.param(False, id="is_sampled-False")]
)
def test_tracestate_propagation(instrument, elasticapm_client, waiting_httpserver, is_sampled):
    traceparent = TraceParent.from_string(
        "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03", "foo=bar,baz=bazzinga"
    )

    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    transaction_object = elasticapm_client.begin_transaction("transaction", trace_parent=traceparent)
    transaction_object.is_sampled = is_sampled
    httplib2.Http().request(url, "GET")
    elasticapm_client.end_transaction("MyView")
    headers = waiting_httpserver.requests[0].headers
    assert headers[constants.TRACESTATE_HEADER_NAME] == "foo=bar,baz=bazzinga"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

import json
import os
import time

from elasticapm import capture_span
from elasticapm.conf import constants
from elasticapm.contrib.serverless.aws import (
    capture_serverless,
    get_data_from_request,
    get_data_from_response,
    should_normalize_headers,
)


@pytest.fixture
def event_api():
    aws_data_file = os.path.join(os.path.dirname(__file__), "aws_api_test_data.json")
    with open(aws_data_file) as f:
        return json.load(f)


@pytest.fixture
def event_api2():
    aws_data_file = os.path.join(os.path.dirname(__file__), "aws_api2_test_data.json")
    with open(aws_data_file) as f:
        return json.load(f)


@pytest.fixture
def event_lurl():
    aws_data_file = os.path.join(os.path.dirname(__file__), "aws_lurl_test_data.json")
    with open(aws_data_file) as f:
        return json.load(f)


@pytest.fixture
def event_elb():
    aws_data_file = os.path.join(os.path.dirname(__file__), "aws_elb_test_data.json")
    with open(aws_data_file) as f:
        return json.load(f)


@pytest.fixture
def event_s3():
    aws_data_file = os.path.join(os.path.dirname(__file__), "aws_s3_test_data.json")
    with open(aws_data_file) as f:
        return json.load(f)


@pytest.fixture
def event_s3_batch():
    aws_data_file = os.path.join(os.path.dirname(__file__), "aws_s3_batch_test_data.json")
    with open(aws_data_file) as f:
        return json.load(f)


@pytest.fixture
def event_sqs():
    aws_data_file = os.path.join(os.path.dirname(__file__), "aws_sqs_test_data.json")
    with open(aws_data_file) as f:
        return json.load(f)


@pytest.fixture
def event_sns():
    aws_data_file = os.path.join(os.path.dirname(__file__), "aws_sns_test_data.json")
    with open(aws_data_file) as f:
        return json.load(f)


@pytest.fixture
def event_list():
    """
    Lambda functions can receive the raw output of Step Functions as the event.
    Because it's any valid JSON, it can be a list as well, with no useful
    context information.
    """
    return ["foo", "bar", "baz"]


@pytest.fixture
def context():
    return SampleContext()


class SampleContext:
    """
    Stand-in for AWS lambda context object
    """

    def __init__(self):
        self.invoked_function_arn = "arn:aws:lambda:us-west-2:123456789012:function:my-function:someAlias"
        self.aws_request_id = "12345"


def test_request_data(event_api, event_api2):
    data = get_data_from_request(event_api, capture_body=True, capture_headers=True)

    assert data["method"] == "GET"
    assert (
        data["url"]["full"]
        == "https://02plqthge2.execute-api.us-east-1.amazonaws.com/dev/fetch_all?test%40key=test%40value"
    )
    assert data["headers"]["Host"] == "02plqthge2.execute-api.us-east-1.amazonaws.com"

    data = get_data_from_request(event_api2, capture_body=True, capture_headers=True)

    assert data["method"] == "GET"
    assert data["url"]["full"] == "https://02plqthge2.execute-api.us-east-1.amazonaws.com/dev/fetch_all"
    assert data["headers"]["host"] == "02plqthge2.execute-api.us-east-1.amazonaws.com"

    data = get_data_from_request(event_api, capture_body=False, capture_headers=False)

    assert data["method"] == "GET"
    assert (
        data["url"]["full"]
        == "https://02plqthge2.execute-api.us-east-1.amazonaws.com/dev/fetch_all?test%40key=test%40value"
    )
    assert "headers" not in data


def test_elb_request_data(event_elb):
    data = get_data_from_request(event_elb, capture_body=True, capture_headers=True)

    assert data["method"] == "POST"
    assert (
        data["url"]["full"]
        == "https://blabla.com/toolz/api/v2.0/downloadPDF/PDF_2020-09-11_11-06-01.pdf?test%40key=test%40value&language=en-DE"
    )
    assert data["headers"]["host"] == "blabla.com"
    assert data["body"] == "blablablabody"

    data = get_data_from_request(event_elb, capture_body=False, capture_headers=False)

    assert data["method"] == "POST"
    assert (
        data["url"]["full"]
        == "https://blabla.com/toolz/api/v2.0/downloadPDF/PDF_2020-09-11_11-06-01.pdf?test%40key=test%40value&language=en-DE"
    )
    assert "headers" not in data
    assert data["body"] == "[REDACTED]"


def test_response_data():
    response = {"statusCode": "200", "headers": {"foo": "bar"}}
    data = get_data_from_response(response, capture_headers=True)
    assert data["status_code"] == 200
    assert data["headers"]["foo"] == "bar"

    response["statusCode"] = 400
    data = get_data_from_response(response, capture_headers=False)
    assert data["status_code"] == 400
    assert "headers" not in data

    data = get_data_from_response({}, capture_headers=False)
    assert not data

    response["statusCode"] = "2xx"
    data = get_data_from_response(response, capture_headers=True)
    assert data["status_code"] == 500


def test_capture_serverless_api_gateway(event_api, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        with capture_span("test_span"):
            time.sleep(0.01)
        return {"statusCode": 200, "headers": {"foo": "bar"}}

    test_func(event_api, context)

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["name"] == "GET /dev/fetch_all"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["span_count"]["started"] == 1
    assert transaction["context"]["request"]["method"] == "GET"
    assert transaction["context"]["request"]["headers"]
    assert transaction["context"]["response"]["status_code"] == 200


def test_capture_serverless_api_gateway_with_args_deprecated(event_api, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    with pytest.warns(PendingDeprecationWarning):

        @capture_serverless(name="test_func")
        def test_func(event, context):
            with capture_span("test_span"):
                time.sleep(0.01)
            return {"statusCode": 200, "headers": {"foo": "bar"}}

        test_func(event_api, context)

        assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
        transaction = elasticapm_client.events[constants.TRANSACTION][0]

        assert transaction["name"] == "GET /dev/fetch_all"
        assert transaction["result"] == "HTTP 2xx"
        assert transaction["span_count"]["started"] == 1
        assert transaction["context"]["request"]["method"] == "GET"
        assert transaction["context"]["request"]["headers"]
        assert transaction["context"]["response"]["status_code"] == 200


def test_capture_serverless_api_gateway_v2(event_api2, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        with capture_span("test_span"):
            time.sleep(0.01)
        return {"statusCode": 200, "headers": {"foo": "bar"}}

    test_func(event_api2, context)

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["name"] == "GET /dev/fetch_all"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["span_count"]["started"] == 1
    assert transaction["context"]["request"]["method"] == "GET"
    assert transaction["context"]["request"]["headers"]
    assert transaction["context"]["response"]["status_code"] == 200
    assert transaction["context"]["cloud"]["origin"]["service"]["name"] == "api gateway"


def test_capture_serverless_lambda_url(event_lurl, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        with capture_span("test_span"):
            time.sleep(0.01)
        return {"statusCode": 200, "headers": {"foo": "bar"}}

    test_func(event_lurl, context)

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["name"] == "GET /dev/fetch_all"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["span_count"]["started"] == 1
    assert transaction["context"]["request"]["method"] == "GET"
    assert transaction["context"]["request"]["headers"]
    assert transaction["context"]["response"]["status_code"] == 200
    assert transaction["context"]["cloud"]["origin"]["service"]["name"] == "lambda url"


def test_capture_serverless_elb(event_elb, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        with capture_span("test_span"):
            time.sleep(0.01)
        return {"statusCode": 200, "headers": {"foo": "bar"}}

    test_func(event_elb, context)

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["name"] == "POST unknown route"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["span_count"]["started"] == 1
    assert transaction["context"]["request"]["method"] == "POST"
    assert transaction["context"]["request"]["headers"]
    assert transaction["context"]["response"]["status_code"] == 200
    assert transaction["context"]["service"]["origin"]["name"] == "lambda-279XGJDqGZ5rsrHC2Fjr"
    assert transaction["trace_id"] == "12345678901234567890123456789012"


def test_capture_serverless_s3(event_s3, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        with capture_span("test_span"):
            time.sleep(0.01)
        return

    test_func(event_s3, context)

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["name"] == "ObjectCreated:Put basepitestbucket"
    assert transaction["span_count"]["started"] == 1


def test_capture_serverless_sns(event_sns, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        with capture_span("test_span"):
            time.sleep(0.01)
        return

    test_func(event_sns, context)

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["name"] == "RECEIVE basepiwstesttopic"
    assert transaction["span_count"]["started"] == 1
    assert transaction["context"]["message"]["headers"]["Population"] == "1250800"
    assert transaction["context"]["message"]["headers"]["City"] == "Any City"
    assert len(transaction["links"]) == 1
    assert transaction["links"][0] == {"trace_id": "0af7651916cd43dd8448eb211c80319c", "span_id": "b7ad6b7169203331"}


def test_capture_serverless_sqs(event_sqs, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        with capture_span("test_span"):
            time.sleep(0.01)
        return

    test_func(event_sqs, context)

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["name"] == "RECEIVE testqueue"
    assert transaction["span_count"]["started"] == 1
    assert transaction["context"]["message"]["headers"]["Population"] == "1250800"
    assert transaction["context"]["message"]["headers"]["City"] == "Any City"
    assert len(transaction["links"]) == 1
    assert transaction["links"][0] == {"trace_id": "0af7651916cd43dd8448eb211c80319c", "span_id": "b7ad6b7169203331"}


def test_capture_serverless_s3_batch(event_s3_batch, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        with capture_span("test_span"):
            time.sleep(0.01)
        return

    test_func(event_s3_batch, context)

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["name"] == "test_func"
    assert transaction["span_count"]["started"] == 1


@pytest.mark.parametrize("elasticapm_client", [{"service_name": "override"}], indirect=True)
def test_service_name_override(event_api, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        with capture_span("test_span"):
            time.sleep(0.01)
        return {"statusCode": 200, "headers": {"foo": "bar"}}

    test_func(event_api, context)

    assert elasticapm_client.build_metadata()["service"]["name"] == "override"


def test_capture_serverless_list_event(event_list, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        return {"statusCode": 200, "headers": {"foo": "bar"}}

    test_func(event_list, context)

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["name"] == "test_func"


def test_partial_transaction(event_api, context, sending_elasticapm_client):
    import elasticapm.contrib.serverless.aws

    elasticapm.contrib.serverless.aws.REGISTER_PARTIAL_TRANSACTIONS = True
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"
    os.environ["ELASTIC_APM_LAMBDA_APM_SERVER"] = "http://localhost:8200"

    @capture_serverless
    def test_func(event, context):
        return {"statusCode": 200, "headers": {"foo": "bar"}}

    test_func(event_api, context)

    assert len(sending_elasticapm_client.httpserver.requests) == 2

    # There should be no spans from the partial transaction
    for payload in sending_elasticapm_client.httpserver.payloads[1]:
        assert "span" not in payload

    request = sending_elasticapm_client.httpserver.requests[0]
    assert request.full_path == "/register/transaction?"
    assert request.content_type == "application/vnd.elastic.apm.transaction+ndjson"
    assert b"metadata" in request.data
    assert b"AWS Lambda" in request.data
    assert b"transaction" in request.data
    sending_elasticapm_client.close()


def test_partial_transaction_failure(event_api, context, sending_elasticapm_client):
    import elasticapm.contrib.serverless.aws

    elasticapm.contrib.serverless.aws.REGISTER_PARTIAL_TRANSACTIONS = True
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"
    os.environ["ELASTIC_APM_LAMBDA_APM_SERVER"] = "http://localhost:8200"
    sending_elasticapm_client.httpserver.code = 404
    sending_elasticapm_client.httpserver.content = "go away"

    @capture_serverless
    def test_func(event, context):
        return {"statusCode": 200, "headers": {"foo": "bar"}}

    test_func(event_api, context)
    test_func(event_api, context)

    assert len(sending_elasticapm_client.httpserver.requests) == 3
    request = sending_elasticapm_client.httpserver.requests[0]
    assert request.full_path == "/register/transaction?"
    assert request.content_type == "application/vnd.elastic.apm.transaction+ndjson"
    assert b"metadata" in request.data
    assert b"AWS Lambda" in request.data
    assert b"transaction" in request.data
    sending_elasticapm_client.close()


def test_with_headers_as_none(event_api2, context, elasticapm_client):
    os.environ["AWS_LAMBDA_FUNCTION_NAME"] = "test_func"

    @capture_serverless
    def test_func(event, context):
        with capture_span("test_span"):
            time.sleep(0.01)
        return {"statusCode": 200, "headers": {"foo": "bar"}}

    event_api2["headers"] = None

    test_func(event_api2, context)
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1


def test_should_normalize_headers_true(event_api, event_elb):
    assert should_normalize_headers(event_api) is True
    assert should_normalize_headers(event_elb) is True


def test_should_normalize_headers_false(event_api2, event_lurl, event_s3, event_s3_batch, event_sqs, event_sns):
    assert should_normalize_headers(event_api2) is False
    assert should_normalize_headers(event_lurl) is False
    assert should_normalize_headers(event_s3) is False
    assert should_normalize_headers(event_s3_batch) is False
    assert should_normalize_headers(event_sqs) is False
    assert should_normalize_headers(event_sns) is False



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import time

import certifi
import mock
import pytest
import urllib3.poolmanager
from urllib3.exceptions import MaxRetryError, TimeoutError

from elasticapm.conf import constants
from elasticapm.transport.exceptions import TransportException
from elasticapm.transport.http import Transport, version_string_to_tuple
from tests.utils import assert_any_record_contains

try:
    import urlparse
except ImportError:
    from urllib import parse as urlparse

CUR_DIR = os.path.dirname(os.path.realpath(__file__))


@pytest.mark.flaky(reruns=3)  # test is flaky on Windows
def test_send(waiting_httpserver, elasticapm_client):
    elasticapm_client.server_version = (8, 0, 0)  # avoid making server_info request
    waiting_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    transport = Transport(
        waiting_httpserver.url, client=elasticapm_client, headers=elasticapm_client._transport._headers
    )
    transport.start_thread()
    try:
        url = transport.send("x".encode("latin-1"))
        assert url == "http://example.com/foo"
        request_headers = waiting_httpserver.requests[0].headers
        assert request_headers["User-Agent"].startswith("apm-agent-python/")
        assert request_headers["Authorization"] == "Bearer test_key"
        assert request_headers["Content-Type"] == "application/x-ndjson"
        assert request_headers["Content-Encoding"] == "gzip"
    finally:
        transport.close()


@mock.patch("urllib3.poolmanager.PoolManager.urlopen")
def test_timeout(mock_urlopen, elasticapm_client):
    elasticapm_client.server_version = (8, 0, 0)  # avoid making server_info request
    transport = Transport("http://localhost", timeout=5, client=elasticapm_client)
    transport.start_thread()
    mock_urlopen.side_effect = MaxRetryError(None, None, reason=TimeoutError())
    try:
        with pytest.raises(TransportException) as exc_info:
            transport.send("x")
        assert "timeout" in str(exc_info.value)
    finally:
        transport.close()


@pytest.mark.flaky(reruns=3)  # test is flaky on Windows
def test_http_error(waiting_httpserver, elasticapm_client):
    elasticapm_client.server_version = (8, 0, 0)  # avoid making server_info request
    waiting_httpserver.serve_content(code=418, content="I'm a teapot")
    transport = Transport(waiting_httpserver.url, client=elasticapm_client)
    transport.start_thread()
    try:
        with pytest.raises(TransportException) as exc_info:
            transport.send("x")
        for val in (418, "I'm a teapot"):
            assert str(val) in str(exc_info.value)
    finally:
        transport.close()


@mock.patch("urllib3.poolmanager.PoolManager.urlopen")
def test_generic_error(mock_urlopen, elasticapm_client):
    url, status, message, body = ("http://localhost:9999", 418, "I'm a teapot", "Nothing")
    elasticapm_client.server_version = (8, 0, 0)  # avoid making server_info request
    transport = Transport(url, client=elasticapm_client)
    transport.start_thread()
    mock_urlopen.side_effect = Exception("Oopsie")
    try:
        with pytest.raises(TransportException) as exc_info:
            transport.send("x")
        assert "Oopsie" in str(exc_info.value)
    finally:
        transport.close()


def test_http_proxy_environment_variable(elasticapm_client):
    with mock.patch.dict("os.environ", {"HTTP_PROXY": "http://example.com"}, clear=True):
        transport = Transport("http://localhost:9999", client=elasticapm_client)
        assert isinstance(transport.http, urllib3.ProxyManager)


def test_https_proxy_environment_variable(elasticapm_client):
    with mock.patch.dict(
        "os.environ",
        {
            "HTTPS_PROXY": "https://example.com",
        },
        clear=True,
    ):
        transport = Transport("http://localhost:9999", client=elasticapm_client)
        assert isinstance(transport.http, urllib3.poolmanager.ProxyManager)


def test_https_proxy_environment_variable_is_preferred(elasticapm_client):
    with mock.patch.dict(
        "os.environ", {"https_proxy": "https://example.com", "HTTP_PROXY": "http://example.com"}, clear=True
    ):
        transport = Transport("http://localhost:9999", client=elasticapm_client)
        assert isinstance(transport.http, urllib3.poolmanager.ProxyManager)
        assert transport.http.proxy.scheme == "https"


def test_no_proxy_star(elasticapm_client):
    with mock.patch.dict("os.environ", {"HTTPS_PROXY": "https://example.com", "NO_PROXY": "*"}, clear=True):
        transport = Transport("http://localhost:9999", client=elasticapm_client)
        assert not isinstance(transport.http, urllib3.poolmanager.ProxyManager)


def test_no_proxy_host(elasticapm_client):
    with mock.patch.dict("os.environ", {"HTTPS_PROXY": "https://example.com", "NO_PROXY": "localhost"}, clear=True):
        transport = Transport("http://localhost:9999", client=elasticapm_client)
        assert not isinstance(transport.http, urllib3.poolmanager.ProxyManager)


def test_no_proxy_all(elasticapm_client):
    with mock.patch.dict("os.environ", {"HTTPS_PROXY": "https://example.com", "NO_PROXY": "*"}, clear=True):
        transport = Transport("http://localhost:9999", client=elasticapm_client)
        assert not isinstance(transport.http, urllib3.poolmanager.ProxyManager)


def test_header_encodings(elasticapm_client):
    """
    Tests that headers are encoded as bytestrings. If they aren't,
    urllib assumes it needs to encode the data as well, which is already a zlib
    encoded bytestring, and explodes.
    """
    headers = {str("X"): str("V")}
    elasticapm_client.server_version = (8, 0, 0)  # avoid making server_info request
    transport = Transport("http://localhost:9999", headers=headers, client=elasticapm_client)
    transport.start_thread()
    try:
        with mock.patch("elasticapm.transport.http.urllib3.PoolManager.urlopen") as mock_urlopen:
            mock_urlopen.return_value = mock.Mock(status=202)
            transport.send("")
        _, args, kwargs = mock_urlopen.mock_calls[0]
        for k, v in kwargs["headers"].items():
            assert isinstance(k, bytes)
            assert isinstance(v, bytes)
    finally:
        transport.close()


@pytest.mark.flaky(reruns=3)  # test is flaky on Windows
def test_ssl_verify_fails(waiting_httpsserver, elasticapm_client):
    waiting_httpsserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    transport = Transport(waiting_httpsserver.url, client=elasticapm_client)
    transport.start_thread()
    try:
        with pytest.raises(TransportException) as exc_info:
            url = transport.send("x".encode("latin-1"))
        assert "certificate verify failed" in str(exc_info)
    finally:
        transport.close()


@pytest.mark.flaky(reruns=3)  # test is flaky on Windows
@pytest.mark.filterwarnings("ignore:Unverified HTTPS")
def test_ssl_verify_disable(waiting_httpsserver, elasticapm_client):
    waiting_httpsserver.serve_content(code=202, content="", headers={"Location": "https://example.com/foo"})
    transport = Transport(waiting_httpsserver.url, verify_server_cert=False, client=elasticapm_client)
    transport.start_thread()
    try:
        url = transport.send("x".encode("latin-1"))
        assert url == "https://example.com/foo"
    finally:
        transport.close()


@pytest.mark.flaky(reruns=3)  # test is flaky on Windows
def test_ssl_verify_disable_http(waiting_httpserver, elasticapm_client):
    """
    Make sure that ``assert_hostname`` isn't passed in for http requests, even
    with verify_server_cert=False
    """
    waiting_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    transport = Transport(waiting_httpserver.url, verify_server_cert=False, client=elasticapm_client)
    transport.start_thread()
    try:
        url = transport.send("x".encode("latin-1"))
        assert url == "http://example.com/foo"
    finally:
        transport.close()


@pytest.mark.flaky(reruns=3)  # test is flaky on Windows
def test_ssl_cert_pinning_http(waiting_httpserver, elasticapm_client):
    """
    Won't fail, as with the other cert pinning test, since certs aren't relevant
    for http, only https.
    """
    waiting_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    transport = Transport(
        waiting_httpserver.url,
        server_cert=os.path.join(os.path.dirname(__file__), "wrong_cert.pem"),
        verify_server_cert=True,
        client=elasticapm_client,
    )
    transport.start_thread()
    try:
        url = transport.send("x".encode("latin-1"))
        assert url == "http://example.com/foo"
    finally:
        transport.close()


@pytest.mark.flaky(reruns=3)  # test is flaky on Windows
def test_ssl_cert_pinning(waiting_httpsserver, elasticapm_client):
    waiting_httpsserver.serve_content(code=202, content="", headers={"Location": "https://example.com/foo"})
    transport = Transport(
        waiting_httpsserver.url,
        server_cert=os.path.join(CUR_DIR, "..", "ca/server.pem"),
        verify_server_cert=True,
        client=elasticapm_client,
    )
    transport.start_thread()
    try:
        url = transport.send("x".encode("latin-1"))
        assert url == "https://example.com/foo"
    finally:
        transport.close()


@pytest.mark.parametrize(
    "sending_elasticapm_client",
    [{"server_ca_cert_file": os.path.normpath(os.path.join(CUR_DIR, "..", "ca/ca.crt"))}],
    indirect=True,
)
def test_custom_ca_cert(sending_elasticapm_client):
    ca_cert = os.path.normpath(os.path.join(CUR_DIR, "..", "ca/ca.crt"))
    pool = sending_elasticapm_client._transport.http
    assert pool.connection_pool_kw["ca_certs"] == ca_cert
    conn = pool.connection_from_url("https://localhost")
    assert conn.ca_certs == ca_cert


@pytest.mark.flaky(reruns=3)  # test is flaky on Windows
def test_ssl_cert_pinning_fails(waiting_httpsserver, elasticapm_client):
    waiting_httpsserver.serve_content(code=202, content="", headers={"Location": "https://example.com/foo"})
    url = waiting_httpsserver.url
    transport = Transport(
        url,
        server_cert=os.path.join(os.path.dirname(__file__), "wrong_cert.pem"),
        verify_server_cert=True,
        client=elasticapm_client,
    )
    transport.start_thread()
    try:
        with pytest.raises(TransportException) as exc_info:
            transport.send("x".encode("latin-1"))
    finally:
        transport.close()

    assert "Fingerprints did not match" in exc_info.value.args[0]


def test_config_url(elasticapm_client):
    transport = Transport("http://example.com/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    assert transport._config_url == "http://example.com/" + constants.AGENT_CONFIG_PATH


def test_get_config(waiting_httpserver, elasticapm_client):
    waiting_httpserver.serve_content(
        code=200, content=b'{"x": "y"}', headers={"Cache-Control": "max-age=5", "Etag": "2"}
    )
    url = waiting_httpserver.url
    transport = Transport(
        url + "/" + constants.EVENTS_API_PATH,
        client=elasticapm_client,
        headers=elasticapm_client._transport._headers,
    )
    version, data, max_age = transport.get_config("1", {})
    assert version == "2"
    assert data == {"x": "y"}
    assert max_age == 5

    assert "Content-Encoding" not in waiting_httpserver.requests[0].headers
    assert waiting_httpserver.requests[0].headers["Content-Type"] == "application/json"


@mock.patch("urllib3.poolmanager.PoolManager.urlopen")
def test_get_config_handle_exception(mock_urlopen, caplog, elasticapm_client):
    transport = Transport("http://example.com/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    mock_urlopen.side_effect = urllib3.exceptions.RequestError(transport.http, "http://example.com/", "boom")
    with caplog.at_level("DEBUG", "elasticapm.transport.http"):
        version, data, max_age = transport.get_config("1", {})
    assert version == "1"
    assert max_age == 300
    record = caplog.records[-1]
    assert "HTTP error" in record.msg


def test_get_config_cache_headers_304(waiting_httpserver, caplog, elasticapm_client):
    waiting_httpserver.serve_content(code=304, content=b"", headers={"Cache-Control": "max-age=5"})
    url = waiting_httpserver.url
    transport = Transport(url + "/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    with caplog.at_level("DEBUG", "elasticapm.transport.http"):
        version, data, max_age = transport.get_config("1", {})
    assert waiting_httpserver.requests[0].headers["If-None-Match"] == "1"
    assert version == "1"
    assert data is None
    assert max_age == 5
    record = caplog.records[-1]
    assert "Configuration unchanged" in record.msg


def test_get_config_bad_cache_control_header(waiting_httpserver, caplog, elasticapm_client):
    waiting_httpserver.serve_content(
        code=200, content=b'{"x": "y"}', headers={"Cache-Control": "max-age=fifty", "Etag": "2"}
    )
    url = waiting_httpserver.url
    transport = Transport(url + "/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    with caplog.at_level("DEBUG", "elasticapm.transport.http"):
        version, data, max_age = transport.get_config("1", {})
    assert version == "2"
    assert data == {"x": "y"}
    assert max_age == 300
    record = caplog.records[-1]
    assert record.message == "Could not parse Cache-Control header: max-age=fifty"


def test_get_config_cache_control_zero(waiting_httpserver, caplog, elasticapm_client):
    waiting_httpserver.serve_content(
        code=200, content=b'{"x": "y"}', headers={"Cache-Control": "max-age=0", "Etag": "2"}
    )
    url = waiting_httpserver.url
    transport = Transport(url + "/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    max_age = transport.get_config("1", {})[2]
    assert max_age == 300  # if max-age is 0, we use the default


def test_get_config_cache_control_negative(waiting_httpserver, caplog, elasticapm_client):
    waiting_httpserver.serve_content(
        code=200, content=b'{"x": "y"}', headers={"Cache-Control": "max-age=-1", "Etag": "2"}
    )
    url = waiting_httpserver.url
    transport = Transport(url + "/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    with caplog.at_level("DEBUG", "elasticapm.transport.http"):
        max_age = transport.get_config("1", {})[2]
    assert max_age == 300  # if max-age is negative, we use the default
    record = caplog.records[-1]
    assert record.message == "Could not parse Cache-Control header: max-age=-1"


def test_get_config_cache_control_less_than_minimum(waiting_httpserver, caplog, elasticapm_client):
    waiting_httpserver.serve_content(
        code=200, content=b'{"x": "y"}', headers={"Cache-Control": "max-age=3", "Etag": "2"}
    )
    url = waiting_httpserver.url
    transport = Transport(url + "/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    max_age = transport.get_config("1", {})[2]
    assert max_age == 5  # if max-age is less than 5, we use 5


def test_get_config_empty_response(waiting_httpserver, caplog, elasticapm_client):
    waiting_httpserver.serve_content(code=200, content=b"", headers={"Cache-Control": "max-age=5"})
    url = waiting_httpserver.url
    transport = Transport(url + "/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    with caplog.at_level("DEBUG", "elasticapm.transport.http"):
        version, data, max_age = transport.get_config("1", {})
    assert version == "1"
    assert data is None
    assert max_age == 5
    record = caplog.records[-1]
    assert record.message == "APM Server answered with empty body and status code 200"


def test_use_certifi(elasticapm_client):
    transport = Transport("/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    assert transport.ca_certs == certifi.where()
    elasticapm_client.config.update("2", use_certifi=False)
    assert not transport.ca_certs


@pytest.mark.parametrize(
    "version,expected",
    [
        (
            "1.2.3",
            (1, 2, 3),
        ),
        (
            "1.2.3-alpha1",
            (1, 2, 3, "alpha1"),
        ),
        (
            "1.2.3alpha1",
            (1, 2, "3alpha1"),
        ),
        (
            "1.2",
            (1, 2),
        ),
    ],
)
def test_server_version_to_tuple(version, expected):
    assert version_string_to_tuple(version) == expected


@pytest.mark.parametrize(
    "version,expected",
    [
        (
            "1.2.3",
            (1, 2, 3),
        ),
        (
            "1.2.3-alpha1",
            (1, 2, 3, "alpha1"),
        ),
        (
            "1.2.3alpha1",
            (1, 2, "3alpha1"),
        ),
        (
            "1.2",
            (1, 2, 0),
        ),
        (
            "8.0.0-alpha1",
            (8, 0, 0, "alpha1"),
        ),
    ],
)
def test_fetch_server_info(waiting_httpserver, elasticapm_client, version, expected):
    waiting_httpserver.serve_content(
        code=200,
        content=f'{{"version": "{version}"}}'.encode("utf-8"),
    )
    url = waiting_httpserver.url
    transport = Transport(
        url + "/" + constants.EVENTS_API_PATH, client=elasticapm_client, headers=elasticapm_client._transport._headers
    )
    transport.fetch_server_info()
    assert elasticapm_client.server_version == expected
    request_headers = waiting_httpserver.requests[0].headers
    assert request_headers["User-Agent"].startswith("apm-agent-python/")
    assert "Authorization" in request_headers
    assert "Content-Type" not in request_headers
    assert "Content-Encoding" not in request_headers


def test_fetch_server_info_no_json(waiting_httpserver, caplog, elasticapm_client):
    waiting_httpserver.serve_content(
        code=200,
        content=b'"version": "8.0.0-alpha1"',
    )
    url = waiting_httpserver.url
    transport = Transport(url + "/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    with caplog.at_level("DEBUG", logger="elasticapm.transport.http"):
        transport.fetch_server_info()
    assert elasticapm_client.server_version is None
    assert_any_record_contains(caplog.records, "JSON decoding error while fetching server information")


def test_fetch_server_info_no_version(waiting_httpserver, caplog, elasticapm_client):
    waiting_httpserver.serve_content(
        code=200,
        content=b"{}",
    )
    url = waiting_httpserver.url
    transport = Transport(url + "/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    with caplog.at_level("DEBUG", logger="elasticapm.transport.http"):
        transport.fetch_server_info()
    assert elasticapm_client.server_version is None
    assert_any_record_contains(caplog.records, "No version key found in server response")


def test_fetch_server_info_flat_string(waiting_httpserver, caplog, elasticapm_client):
    waiting_httpserver.serve_content(
        code=200,
        content=b'"8.0.0-alpha1"',
    )
    url = waiting_httpserver.url
    transport = Transport(url + "/" + constants.EVENTS_API_PATH, client=elasticapm_client)
    with caplog.at_level("DEBUG", logger="elasticapm.transport.http"):
        transport.fetch_server_info()
    assert elasticapm_client.server_version is None
    assert_any_record_contains(caplog.records, "No version key found in server response")


def test_skip_server_info(waiting_httpserver, elasticapm_client):
    elasticapm_client.config.update(version="1", skip_server_info=True)
    waiting_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    transport = Transport(
        waiting_httpserver.url, client=elasticapm_client, headers=elasticapm_client._transport._headers
    )
    transport.start_thread()
    try:
        url = transport.send("x".encode("latin-1"))
        assert url == "http://example.com/foo"
    finally:
        transport.close()

    assert elasticapm_client.server_version is None
    assert elasticapm_client.check_server_version(gte=(8, 7, 1))


def test_close(waiting_httpserver, elasticapm_client):
    elasticapm_client.server_version = (8, 0, 0)  # avoid making server_info request
    waiting_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    transport = Transport(
        waiting_httpserver.url, client=elasticapm_client, headers=elasticapm_client._transport._headers
    )
    transport.start_thread()

    transport.close()

    assert transport._closed is True
    assert transport._flushed.is_set() is True


def test_close_does_nothing_if_called_from_another_pid(waiting_httpserver, caplog, elasticapm_client):
    elasticapm_client.server_version = (8, 0, 0)  # avoid making server_info request
    waiting_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    transport = Transport(
        waiting_httpserver.url, client=elasticapm_client, headers=elasticapm_client._transport._headers
    )
    transport.start_thread()

    with mock.patch("os.getpid") as getpid_mock:
        getpid_mock.return_value = 0
        transport.close()

    assert transport._closed is False

    transport.close()


def test_close_can_be_called_multiple_times(waiting_httpserver, caplog, elasticapm_client):
    elasticapm_client.server_version = (8, 0, 0)  # avoid making server_info request
    waiting_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    transport = Transport(
        waiting_httpserver.url, client=elasticapm_client, headers=elasticapm_client._transport._headers
    )
    transport.start_thread()

    with caplog.at_level("INFO", logger="elasticapm.transport.http"):
        transport.close()

    assert transport._closed is True

    transport.close()


def test_close_timeout_error_without_flushing(waiting_httpserver, caplog, elasticapm_client):
    elasticapm_client.server_version = (8, 0, 0)  # avoid making server_info request
    waiting_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})

    with caplog.at_level("INFO", logger="elasticapm.transport.http"):
        with mock.patch.object(Transport, "_max_flush_time_seconds", 0):
            with mock.patch.object(Transport, "_flush") as flush_mock:
                # sleep more that the timeout
                flush_mock.side_effect = lambda x: time.sleep(0.1)
                transport = Transport(
                    waiting_httpserver.url, client=elasticapm_client, headers=elasticapm_client._transport._headers
                )
                transport.start_thread()
                # need to write something to the buffer to have _flush() called
                transport.queue("error", {"an": "error"})
                transport.close()

    assert transport._flushed.is_set() is False
    assert transport._closed is True
    record = caplog.records[-1]
    assert "Closing the transport connection timed out." in record.msg


def test_http_pool_manager_is_recycled_at_stop_thread(waiting_httpserver, caplog, elasticapm_client):
    elasticapm_client.server_version = (8, 0, 0)  # avoid making server_info request
    waiting_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    transport = Transport(
        waiting_httpserver.url, client=elasticapm_client, headers=elasticapm_client._transport._headers
    )
    transport.start_thread()
    pool_manager = transport.http

    with caplog.at_level("INFO", logger="elasticapm.transport.http"):
        transport.stop_thread()

    assert transport._flushed.is_set() is True
    assert pool_manager != transport._http
    assert not caplog.records



# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

import mock
import pytest
import wrapt

import elasticapm
from elasticapm.conf import constants
from elasticapm.conf.constants import SPAN, TRANSACTION
from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from tests.utils import assert_any_record_contains


class Dummy(object):
    def dummy(self, call_args=None):
        return call_args


class _TestInstrumentNonExistingFunctionOnModule(AbstractInstrumentedModule):
    name = "test_non_existing_function_instrumentation"
    instrument_list = [("os.path", "non_existing_function")]


class _TestInstrumentNonExistingMethod(AbstractInstrumentedModule):
    name = "test_non_existing_method_instrumentation"
    instrument_list = [("logging", "Logger.non_existing_method")]


class _TestInstrumentBrokenModule(AbstractInstrumentedModule):
    name = "test_non_existing_method_instrumentation"
    instrument_list = [("tests.instrumentation.broken_class", "C.foo")]


class _TestDummyInstrumentation(AbstractInstrumentedModule):
    name = "test_dummy_instrument"

    instrument_list = [("tests.instrumentation.base_tests", "Dummy.dummy")]

    def call(self, module, method, wrapped, instance, args, kwargs):
        kwargs = kwargs or {}
        kwargs["call_args"] = (module, method)
        with elasticapm.capture_span("dummy"):
            return wrapped(*args, **kwargs)


def test_instrument_nonexisting_method_on_module():
    _TestInstrumentNonExistingFunctionOnModule().instrument()


def test_instrument_nonexisting_method(caplog):
    with caplog.at_level(logging.DEBUG, "elasticapm.instrument"):
        _TestInstrumentNonExistingMethod().instrument()
    assert_any_record_contains(caplog.records, "has no attribute", "elasticapm.instrument")


def test_instrument_broken_module(caplog):
    with caplog.at_level(logging.DEBUG, "elasticapm.instrument"):
        _TestInstrumentBrokenModule().instrument()
    assert_any_record_contains(caplog.records, "due to unknown error", "elasticapm.instrument")


def test_double_instrument(elasticapm_client):
    elasticapm_client.begin_transaction("test")
    inst = _TestDummyInstrumentation()
    try:
        inst.instrument()
        assert hasattr(Dummy.dummy, "_self_wrapper")
        Dummy().dummy()
        elasticapm_client.end_transaction()
        assert len(elasticapm_client.spans_for_transaction(elasticapm_client.events[constants.TRANSACTION][0])) == 1
        inst.instrumented = False
        inst.instrument()
        elasticapm_client.begin_transaction("test")
        Dummy().dummy()
        elasticapm_client.end_transaction()
        assert len(elasticapm_client.spans_for_transaction(elasticapm_client.events[constants.TRANSACTION][1])) == 1
    finally:
        inst.uninstrument()


def test_uninstrument(caplog):
    original = Dummy.dummy
    assert not isinstance(Dummy.dummy, wrapt.BoundFunctionWrapper)

    instrumentation = _TestDummyInstrumentation()
    with caplog.at_level(logging.DEBUG, "elasticapm.instrument"):
        instrumentation.instrument()
    assert_any_record_contains(
        caplog.records,
        "Instrumented test_dummy_instrument, tests.instrumentation.base_tests.Dummy.dummy",
        "elasticapm.instrument",
    )
    assert Dummy.dummy is not original
    assert isinstance(Dummy.dummy, wrapt.BoundFunctionWrapper)

    with caplog.at_level(logging.DEBUG, "elasticapm.instrument"):
        instrumentation.uninstrument()
    assert_any_record_contains(
        caplog.records,
        "Uninstrumented test_dummy_instrument, tests.instrumentation.base_tests.Dummy.dummy",
        "elasticapm.instrument",
    )
    assert Dummy.dummy is original
    assert not isinstance(Dummy.dummy, wrapt.BoundFunctionWrapper)


def test_module_method_args(elasticapm_client):
    """
    Test that the module/method arguments are correctly passed to
    the _TestDummyInstrumentation.call method
    """
    instrumentation = _TestDummyInstrumentation()
    instrumentation.instrument()
    elasticapm_client.begin_transaction("test")
    dummy = Dummy()
    call_args = dummy.dummy()
    elasticapm_client.end_transaction("test", "test")
    instrumentation.uninstrument()

    assert call_args == ("tests.instrumentation.base_tests", "Dummy.dummy")


def test_skip_instrument_env_var(caplog):
    instrumentation = _TestDummyInstrumentation()
    with mock.patch.dict("os.environ", {"SKIP_INSTRUMENT_TEST_DUMMY_INSTRUMENT": "foo"}), caplog.at_level(
        logging.DEBUG, "elasticapm.instrument"
    ):
        instrumentation.instrument()
    assert_any_record_contains(caplog.records, "Skipping", "elasticapm.instrument")
    assert not instrumentation.instrumented


def test_skip_ignored_frames(elasticapm_client):
    elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span("test"):
        pass
    elasticapm_client.end_transaction("test", "test")
    span = elasticapm_client.events[SPAN][0]
    for frame in span["stacktrace"]:
        assert not frame["module"].startswith("elasticapm")


def test_end_nonexisting_span(caplog, elasticapm_client):
    with caplog.at_level(logging.DEBUG, "elasticapm.traces"):
        t = elasticapm_client.begin_transaction("test")
        # we're purposefully creating a case where we don't begin a span
        # and then try to end the non-existing span
        t.is_sampled = False
        with elasticapm.capture_span("test_name", "test_type"):
            t.is_sampled = True
    elasticapm_client.end_transaction("test", "")
    assert_any_record_contains(
        caplog.records, "ended non-existing span test_name of type test_type", "elasticapm.traces"
    )


def test_outcome_by_span_exception(elasticapm_client):
    elasticapm_client.begin_transaction("test")
    try:
        with elasticapm.capture_span("fail", "test_type"):
            assert False
    except AssertionError:
        pass
    with elasticapm.capture_span("success", "test_type"):
        pass
    elasticapm_client.end_transaction("test")
    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    assert spans[0]["name"] == "fail" and spans[0]["outcome"] == "failure"
    assert spans[1]["name"] == "success" and spans[1]["outcome"] == "success"


@pytest.mark.parametrize(
    "outcome,http_status_code,log_message,result",
    [
        (None, 200, None, "success"),
        (None, 500, None, "failure"),
        (None, "500", None, "failure"),
        (None, "HTTP 500", "Invalid HTTP status 'HTTP 500' provided", "unknown"),
        ("failure", 200, None, "failure"),  # explicit outcome has precedence
        ("failed", None, "Invalid outcome 'failed' provided", "unknown"),
    ],
)
def test_transaction_outcome(elasticapm_client, caplog, outcome, http_status_code, log_message, result):
    transaction = elasticapm_client.begin_transaction("test")
    with caplog.at_level(logging.INFO, "elasticapm.traces"):
        elasticapm.set_transaction_outcome(outcome=outcome, http_status_code=http_status_code)
    assert transaction.outcome == result
    if log_message is None:
        assert not [True for record in caplog.records if record.name == "elasticapm.traces"]
    else:
        assert_any_record_contains(caplog.records, log_message, "elasticapm.traces")


def test_transaction_outcome_override(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    elasticapm.set_transaction_outcome(constants.OUTCOME.FAILURE)

    assert transaction.outcome == constants.OUTCOME.FAILURE

    elasticapm.set_transaction_outcome(constants.OUTCOME.SUCCESS, override=False)
    # still a failure
    assert transaction.outcome == constants.OUTCOME.FAILURE

    elasticapm.set_transaction_outcome(constants.OUTCOME.SUCCESS, override=True)
    assert transaction.outcome == constants.OUTCOME.SUCCESS


def test_empty_span_name(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span():
        pass
    elasticapm_client.end_transaction("test")

    assert elasticapm_client.events[SPAN][0]["name"] == "unnamed"



# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from elasticapm.instrumentation.packages.mysql import extract_signature


def test_insert():
    sql = """INSERT INTO `mytable` (id, name) VALUE ('2323', 'Ron')"""
    actual = extract_signature(sql)

    assert "INSERT INTO mytable" == actual


def test_update():
    sql = """UPDATE `mytable` set name='Ron' WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "UPDATE mytable" == actual


def test_delete():
    sql = """DELETE FROM `mytable` WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "DELETE FROM mytable" == actual


def test_select_simple():
    sql = """SELECT `id`, `name` FROM `mytable` WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_with_entity_quotes():
    sql = """SELECT `id`, `name` FROM `mytable` WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_with_difficult_values():
    sql = """SELECT id, 'some \\'name' + " from Denmark" FROM `mytable` WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_with_difficult_table_name():
    sql = """SELECT id FROM `myta\n-æøåble` WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM myta\n-æøåble" == actual


def test_select_subselect():
    sql = """SELECT id, name FROM (
            SELECT id, "not a FROM ''value" FROM mytable WHERE id = 2323
    ) LIMIT 20"""
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_subselect_with_alias():
    sql = """
    SELECT count(*)
    FROM (
        SELECT count(id) AS some_alias, some_column
        FROM mytable
        GROUP BY some_colun
        HAVING count(id) > 1
    ) AS foo
    """
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_with_multiple_tables():
    sql = """SELECT count(table2.id)
        FROM table1, table2, table2
        WHERE table2.id = table1.table2_id
    """
    actual = extract_signature(sql)
    assert "SELECT FROM table1" == actual


def test_select_with_invalid_literal():
    sql = 'SELECT "neverending literal FROM (SELECT * FROM ...' ""
    actual = extract_signature(sql)

    assert "SELECT FROM" == actual


def test_savepoint():
    sql = """SAVEPOINT x_asd1234"""
    actual = extract_signature(sql)

    assert "SAVEPOINT" == actual


def test_begin():
    sql = """BEGIN"""
    actual = extract_signature(sql)

    assert "BEGIN" == actual


def test_create_index_with_name():
    sql = """CREATE INDEX myindex ON mytable"""
    actual = extract_signature(sql)

    assert "CREATE INDEX" == actual


def test_create_index_without_name():
    sql = """CREATE INDEX ON mytable"""
    actual = extract_signature(sql)

    assert "CREATE INDEX" == actual


def test_drop_table():
    sql = """DROP TABLE mytable"""
    actual = extract_signature(sql)

    assert "DROP TABLE" == actual


def test_multi_statement_sql():
    sql = """CREATE TABLE mytable; SELECT * FROM mytable; DROP TABLE mytable"""
    actual = extract_signature(sql)

    assert "CREATE TABLE" == actual



#  BSD 3-Clause License
#
#  Copyright (c) 2023, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

pytest.importorskip("starlette")  # isort:skip

from starlette.applications import Starlette
from starlette.responses import PlainTextResponse
from starlette.routing import Route
from starlette.testclient import TestClient

from elasticapm import async_capture_span
from elasticapm.conf import constants
from elasticapm.contrib.starlette import ElasticAPM


def test_starlette_app_instrumented(wrapper_instrument, elasticapm_client):
    async def hi(request):
        body = await request.body()
        with async_capture_span("test"):
            pass
        return PlainTextResponse(str(len(body)))

    routes = [Route("/", hi)]
    app = Starlette(routes=routes)
    test_client = TestClient(app)

    app.build_middleware_stack()
    assert any(middleware.cls is ElasticAPM for middleware in app.user_middleware)
    test_client.get("/")
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["name"] == "GET /"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["outcome"] == "success"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import time

import pytest
from flask import Flask, Response, make_response, render_template, signals
from pytest_localserver.http import WSGIServer

import elasticapm
from elasticapm.contrib.flask import ElasticAPM


@pytest.fixture()
def flask_app():
    app = Flask(__name__)

    @app.route("/an-error/", methods=["GET", "POST"])
    def an_error():
        raise ValueError("hello world")

    @app.route("/users/", methods=["GET", "POST"])
    def users():
        response = make_response(render_template("users.html", users=["Ron", "Rasmus"]))
        response.headers.add("foo", "bar")
        response.headers.add("foo", "baz")
        response.headers.add("bar", "bazzinga")
        return response

    @app.route("/non-standard-status/", methods=["GET", "POST"])
    def non_standard_status():
        return "foo", "fail"

    @app.route("/streaming/", methods=["GET"])
    def streaming():
        def generator():
            for i in range(5):
                with elasticapm.capture_span("generator"):
                    time.sleep(0.01)
                    yield str(i)

        return Response(generator(), mimetype="text/plain")

    @app.route("/transaction-name/", methods=["GET"])
    def transaction_name():
        elasticapm.set_transaction_name("foo")
        elasticapm.set_transaction_result("okydoky")
        return Response("")

    return app


@pytest.fixture()
def flask_wsgi_server(request, flask_app, elasticapm_client):
    server = WSGIServer(application=flask_app)
    apm_client = ElasticAPM(app=flask_app, client=elasticapm_client)
    flask_app.apm_client = apm_client
    server.start()
    try:
        yield server
    finally:
        elasticapm.uninstrument()
        server.stop()
        apm_client.client.close()


@pytest.fixture()
def flask_apm_client(request, flask_app, elasticapm_client):
    client_config = getattr(request, "param", {})
    client_config.setdefault("app", flask_app)
    client_config.setdefault("client", elasticapm_client)
    client = ElasticAPM(**client_config)
    try:
        yield client
    finally:
        elasticapm.uninstrument()
        signals.request_started.disconnect(client.request_started)
        signals.request_finished.disconnect(client.request_finished)
        # remove logging handler if it was added
        logger = logging.getLogger()
        for handler in list(logger.handlers):
            if getattr(handler, "client", None) is client.client:
                logger.removeHandler(handler)


@pytest.fixture()
def sending_flask_apm_client(request, flask_app, sending_elasticapm_client):
    client_config = getattr(request, "param", {})
    client_config.setdefault("app", flask_app)
    client_config.setdefault("client", sending_elasticapm_client)
    client = ElasticAPM(**client_config)
    try:
        yield client
    finally:
        elasticapm.uninstrument()
        signals.request_started.disconnect(client.request_started)
        signals.request_finished.disconnect(client.request_finished)
        # remove logging handler if it was added
        logger = logging.getLogger()
        for handler in list(logger.handlers):
            if getattr(handler, "client", None) is client.client:
                logger.removeHandler(handler)


@pytest.fixture()
def flask_celery(flask_apm_client):
    from celery import Celery

    flask_app = flask_apm_client.app
    celery = Celery(flask_app.import_name, backend=None, broker=None)
    celery.conf.update(CELERY_ALWAYS_EAGER=True)
    TaskBase = celery.Task

    class ContextTask(TaskBase):
        abstract = True

        def __call__(self, *args, **kwargs):
            with flask_app.app_context():
                return TaskBase.__call__(self, *args, **kwargs)

    celery.Task = ContextTask
    celery.flask_apm_client = flask_apm_client
    return celery



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

flask = pytest.importorskip("flask")  # isort:skip
celery = pytest.importorskip("celery")  # isort:skip

import mock

from elasticapm.conf.constants import ERROR, TRANSACTION

pytestmark = [pytest.mark.celery, pytest.mark.flask]


def test_task_failure(flask_celery):
    apm_client = flask_celery.flask_apm_client.client

    @flask_celery.task()
    def failing_task():
        raise ValueError("foo")

    t = failing_task.delay()
    assert t.status == "FAILURE"
    assert len(apm_client.events[ERROR]) == 1
    error = apm_client.events[ERROR][0]
    assert error["culprit"] == "tests.contrib.celery.flask_tests.failing_task"
    assert error["exception"]["message"] == "ValueError: foo"
    assert error["exception"]["handled"] is False

    transaction = apm_client.events[TRANSACTION][0]
    assert transaction["name"] == "tests.contrib.celery.flask_tests.failing_task"
    assert transaction["type"] == "celery"
    assert transaction["result"] == "FAILURE"
    assert transaction["outcome"] == "failure"


def test_task_instrumentation(flask_celery):
    apm_client = flask_celery.flask_apm_client.client

    @flask_celery.task()
    def successful_task():
        return "OK"

    t = successful_task.delay()

    assert t.status == "SUCCESS"
    assert len(apm_client.events[TRANSACTION]) == 1
    transaction = apm_client.events[TRANSACTION][0]
    assert transaction["name"] == "tests.contrib.celery.flask_tests.successful_task"
    assert transaction["type"] == "celery"
    assert transaction["result"] == "SUCCESS"
    assert transaction["outcome"] == "success"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import gzip
import platform
import random
import string
import sys
import time
import timeit

import mock
import pytest

from elasticapm.transport.base import Transport, TransportState
from elasticapm.transport.exceptions import TransportException
from tests.fixtures import DummyTransport, TempStoreClient
from tests.utils import assert_any_record_contains


def test_transport_state_should_try_online():
    state = TransportState()
    assert state.should_try() is True


def test_transport_state_should_try_new_error():
    state = TransportState()
    state.status = state.ERROR
    state.last_check = timeit.default_timer()
    state.retry_number = 1
    assert state.should_try() is False


def test_transport_state_should_try_time_passed_error():
    state = TransportState()
    state.status = state.ERROR
    state.last_check = timeit.default_timer() - 10
    state.retry_number = 1
    assert state.should_try() is True


def test_transport_state_set_fail():
    state = TransportState()
    state.set_fail()
    assert state.status == state.ERROR
    assert state.last_check is not None
    assert state.retry_number == 0


def test_transport_state_set_success():
    state = TransportState()
    state.status = state.ERROR
    state.last_check = "foo"
    state.retry_number = 5
    state.set_success()
    assert state.status == state.ONLINE
    assert state.last_check is None
    assert state.retry_number == -1


@mock.patch("elasticapm.transport.base.Transport.send")
@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {"api_request_time": "5s", "server_url": "http://localhost:8200"},
        {"api_request_time": "5s", "server_url": "http://remotehost:8200"},
    ],
    indirect=True,
)
def test_empty_queue_flush(mock_send, elasticapm_client):
    transport = Transport(client=elasticapm_client)
    try:
        transport.start_thread()
        transport.flush()
        if "localhost:" in elasticapm_client.config.server_url:
            assert mock_send.call_count == 1
        else:
            assert mock_send.call_count == 0
    finally:
        transport.close()


@mock.patch("elasticapm.transport.base.Transport._flush")
@pytest.mark.parametrize("elasticapm_client", [{"api_request_time": "5s"}], indirect=True)
def test_metadata_prepended(mock_flush, elasticapm_client):
    transport = Transport(client=elasticapm_client, compress_level=0)
    transport.start_thread()
    transport.queue("error", {}, flush=True)
    transport.close()
    assert mock_flush.call_count == 1
    args, kwargs = mock_flush.call_args
    buffer = args[0]
    # this test used to mock send but after we fixed a leak for not releasing the memoryview containing
    # the gzipped data we cannot read it anymore. So reimplement _flush and read the data ourselves
    fileobj = buffer.fileobj
    buffer.close()
    compressed_data = fileobj.getbuffer()
    data = gzip.decompress(compressed_data)
    data = data.decode("utf-8").split("\n")
    assert "metadata" in data[0]
    compressed_data.release()


@mock.patch("elasticapm.transport.base.Transport.send")
@pytest.mark.parametrize("elasticapm_client", [{"api_request_time": "100ms"}], indirect=True)
def test_flush_time(mock_send, caplog, elasticapm_client):
    with caplog.at_level("DEBUG", "elasticapm.transport"):
        transport = Transport(client=elasticapm_client)
        transport.start_thread()
        # let first run finish
        time.sleep(0.2)
        transport.close()
    assert_any_record_contains(caplog.records, "due to time since last flush", "elasticapm.transport")
    assert mock_send.call_count == 0


@pytest.mark.flaky(reruns=3)  # test is flaky on Windows
@mock.patch("elasticapm.transport.base.Transport.send")
def test_api_request_time_dynamic(mock_send, caplog, elasticapm_client):
    elasticapm_client.config.update(version="1", api_request_time="1s")
    with caplog.at_level("DEBUG", "elasticapm.transport"):
        transport = Transport(client=elasticapm_client)
        transport.start_thread()
        # let first run finish
        time.sleep(0.2)
        transport.close()
    assert not caplog.records
    assert mock_send.call_count == 0
    elasticapm_client.config.update(version="1", api_request_time="100ms")
    with caplog.at_level("DEBUG", "elasticapm.transport"):
        transport = Transport(client=elasticapm_client)
        transport.start_thread()
        # let first run finish
        time.sleep(0.2)
        transport.close()
    assert_any_record_contains(caplog.records, "due to time since last flush", "elasticapm.transport")
    assert mock_send.call_count == 0


def _cleanup_flush_mock_buffers(mock_flush):
    args, kwargs = mock_flush.call_args
    buffer = args[0]
    buffer.close()


@mock.patch("elasticapm.transport.base.Transport._flush")
def test_api_request_size_dynamic(mock_flush, caplog, elasticapm_client):
    elasticapm_client.config.update(version="1", api_request_size="9b")
    transport = Transport(client=elasticapm_client, queue_chill_count=1)
    transport.start_thread()
    try:
        with caplog.at_level("DEBUG", "elasticapm.transport"):
            transport.queue("error", "".join(random.choice(string.ascii_letters) for i in range(2000)))
            transport._flushed.wait(timeout=0.1)
        _cleanup_flush_mock_buffers(mock_flush)
        assert mock_flush.call_count == 1
        elasticapm_client.config.update(version="1", api_request_size="1mb")
        with caplog.at_level("DEBUG", "elasticapm.transport"):
            transport.queue("error", "".join(random.choice(string.ascii_letters) for i in range(2000)))
            transport._flushed.wait(timeout=0.1)
        # Should be unchanged because our buffer limit is much higher.
        assert mock_flush.call_count == 1
    finally:
        transport.close()


@mock.patch("elasticapm.transport.base.Transport._flush")
@pytest.mark.parametrize("elasticapm_client", [{"api_request_size": "9b"}], indirect=True)
def test_flush_time_size(mock_flush, caplog, elasticapm_client):
    transport = Transport(client=elasticapm_client, queue_chill_count=1)
    transport.start_thread()
    try:
        with caplog.at_level("DEBUG", "elasticapm.transport"):
            transport.queue("error", "".join(random.choice(string.ascii_letters) for i in range(2000)))
            transport._flushed.wait(timeout=0.1)
        _cleanup_flush_mock_buffers(mock_flush)
        assert mock_flush.call_count == 1
    finally:
        transport.close()


@mock.patch("elasticapm.transport.base.Transport.send")
@pytest.mark.parametrize("elasticapm_client", [{"api_request_size": "1000b"}], indirect=True)
def test_forced_flush(mock_send, caplog, elasticapm_client):
    transport = Transport(client=elasticapm_client, compress_level=0)
    transport.start_thread()
    try:
        with caplog.at_level("DEBUG", "elasticapm.transport"):
            transport.queue("error", "x", flush=True)
    finally:
        transport.close()
    assert mock_send.call_count == 1
    assert transport._queued_data is None


@mock.patch("elasticapm.transport.base.Transport.send")
def test_sync_transport_fail_and_recover(mock_send, caplog):
    transport = Transport(client=None)
    transport.start_thread()
    try:
        mock_send.side_effect = TransportException("meh")
        transport.queue("x", {})
        transport.flush()
        assert transport.state.did_fail()
        # first retry should be allowed immediately
        assert transport.state.should_try()

        # recover
        mock_send.side_effect = None
        transport.queue("x", {})
        transport.flush()
        assert not transport.state.did_fail()
    finally:
        transport.close()


@pytest.mark.parametrize("sending_elasticapm_client", [{"api_request_time": "2s"}], indirect=True)
def test_send_timer(sending_elasticapm_client, caplog):
    with caplog.at_level("DEBUG", "elasticapm.transport"):
        assert sending_elasticapm_client.config.api_request_time.total_seconds() == 2
        sending_elasticapm_client.begin_transaction("test_type")
        sending_elasticapm_client.end_transaction("test")

        sending_elasticapm_client._transport.flush()

    assert_any_record_contains(caplog.records, "Sent request")


def test_compress_level_sanitization():
    assert DummyTransport(compress_level=None, url="", client=None)._compress_level == 0
    assert DummyTransport(compress_level=-1, url="", client=None)._compress_level == 0
    assert DummyTransport(compress_level=10, url="", client=None)._compress_level == 9


@mock.patch("elasticapm.transport.base.Transport.send")
def test_transport_metadata_pid_change(mock_send, elasticapm_client):
    transport = Transport(client=elasticapm_client)
    assert not transport._metadata
    transport.start_thread()
    time.sleep(0.2)
    assert transport._metadata
    transport.close()


def test_flushed_arg(sending_elasticapm_client):
    sending_elasticapm_client.begin_transaction("test_type")
    sending_elasticapm_client.end_transaction("test")
    sending_elasticapm_client._transport.flush()

    assert sending_elasticapm_client.httpserver.requests[0].args["flushed"] == "true"


@pytest.mark.skipif(platform.system() == "Windows", reason="Flaky test on windows")
@pytest.mark.flaky(reruns=3)  # Trying to test automatic flushes is inherently flaky
@pytest.mark.parametrize("sending_elasticapm_client", [{"api_request_time": "100ms"}], indirect=True)
def test_flushed_arg_with_wait(sending_elasticapm_client):
    sending_elasticapm_client.begin_transaction("test_type")
    sending_elasticapm_client.end_transaction("test")
    time.sleep(0.3)
    sending_elasticapm_client._transport.flush()

    assert sending_elasticapm_client.httpserver.requests[1].args["flushed"] == "true"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

aiohttp = pytest.importorskip("aiohttp")  # isort:skip
yarl = pytest.importorskip("yarl")  # isort:skip

from elasticapm.conf import constants
from elasticapm.utils.disttracing import TraceParent

pytestmark = [pytest.mark.asyncio, pytest.mark.aiohttp]


@pytest.mark.parametrize("use_yarl", [True, False])
async def test_http_get(instrument, event_loop, elasticapm_client, waiting_httpserver, use_yarl):
    assert event_loop.is_running()
    elasticapm_client.begin_transaction("test")

    url = waiting_httpserver.url
    url = yarl.URL(url) if use_yarl else url

    async with aiohttp.ClientSession() as session:
        async with session.get(waiting_httpserver.url) as resp:
            status = resp.status
            text = await resp.text()

    elasticapm_client.end_transaction()
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1
    span = spans[0]
    assert span["name"] == "GET %s:%s" % waiting_httpserver.server_address
    assert span["type"] == "external"
    assert span["subtype"] == "http"
    assert span["sync"] is False
    assert span["context"]["http"]["url"] == waiting_httpserver.url
    assert span["context"]["http"]["status_code"] == 204
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % waiting_httpserver.server_address[1],
        "type": "",
    }
    assert spans[0]["context"]["service"]["target"]["type"] == "http"
    assert spans[0]["context"]["service"]["target"]["name"] == f"127.0.0.1:{waiting_httpserver.server_address[1]}"
    assert spans[0]["outcome"] == "success"


@pytest.mark.parametrize("use_yarl", [True, False])
async def test_http_get_via_request(instrument, event_loop, elasticapm_client, waiting_httpserver, use_yarl):
    assert event_loop.is_running()
    elasticapm_client.begin_transaction("test")

    url = waiting_httpserver.url
    url = yarl.URL(url) if use_yarl else url

    async with aiohttp.ClientSession() as session:
        await session._request("GET", str_or_url=waiting_httpserver.url)

    elasticapm_client.end_transaction()
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1
    span = spans[0]
    assert span["name"] == "GET %s:%s" % waiting_httpserver.server_address
    assert span["type"] == "external"
    assert span["subtype"] == "http"
    assert span["sync"] is False
    assert span["context"]["http"]["url"] == waiting_httpserver.url
    assert span["context"]["http"]["status_code"] == 204
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % waiting_httpserver.server_address[1],
        "type": "",
    }
    assert spans[0]["context"]["service"]["target"]["type"] == "http"
    assert spans[0]["context"]["service"]["target"]["name"] == f"127.0.0.1:{waiting_httpserver.server_address[1]}"
    assert spans[0]["outcome"] == "success"


@pytest.mark.parametrize("status_code", [400, 500])
async def test_http_get_error(instrument, event_loop, elasticapm_client, waiting_httpserver, status_code):
    assert event_loop.is_running()
    elasticapm_client.begin_transaction("test")
    waiting_httpserver.serve_content("", code=status_code)
    url = waiting_httpserver.url

    async with aiohttp.ClientSession() as session:
        async with session.get(waiting_httpserver.url) as resp:
            status = resp.status
            text = await resp.text()

    elasticapm_client.end_transaction()
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1
    span = spans[0]
    assert span["name"] == "GET %s:%s" % waiting_httpserver.server_address
    assert span["type"] == "external"
    assert span["subtype"] == "http"
    assert span["sync"] is False
    assert span["context"]["http"]["url"] == waiting_httpserver.url
    assert span["context"]["http"]["status_code"] == status_code
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % waiting_httpserver.server_address[1],
        "type": "",
    }
    assert spans[0]["outcome"] == "failure"


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        pytest.param({"use_elastic_traceparent_header": True}, id="use_elastic_traceparent_header-True"),
        pytest.param({"use_elastic_traceparent_header": False}, id="use_elastic_traceparent_header-False"),
    ],
    indirect=True,
)
async def test_trace_parent_propagation_sampled(instrument, event_loop, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    elasticapm_client.begin_transaction("transaction")
    async with aiohttp.ClientSession() as session:
        async with session.get(waiting_httpserver.url) as resp:
            status = resp.status
            text = await resp.text()
    elasticapm_client.end_transaction("MyView")
    transactions = elasticapm_client.events[constants.TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    headers = waiting_httpserver.requests[0].headers
    assert constants.TRACEPARENT_HEADER_NAME in headers
    trace_parent = TraceParent.from_string(
        headers[constants.TRACEPARENT_HEADER_NAME], tracestate_string=headers[constants.TRACESTATE_HEADER_NAME]
    )
    assert trace_parent.trace_id == transactions[0]["trace_id"]
    assert trace_parent.span_id == spans[0]["id"]
    assert trace_parent.trace_options.recorded
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict

    if elasticapm_client.config.use_elastic_traceparent_header:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME in headers
        assert headers[constants.TRACEPARENT_HEADER_NAME] == headers[constants.TRACEPARENT_LEGACY_HEADER_NAME]
    else:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME not in headers


@pytest.mark.parametrize("sampled", [True, False])
async def test_trace_parent_propagation_sampled_headers_none(
    instrument, event_loop, elasticapm_client, waiting_httpserver, sampled
):
    """
    Test that we don't blow up if headers are explicitly set to None
    """
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    transaction = elasticapm_client.begin_transaction("transaction")
    transaction.is_sampled = sampled
    async with aiohttp.ClientSession() as session:
        async with session.get(waiting_httpserver.url, headers=None) as resp:
            status = resp.status
            text = await resp.text()
    elasticapm_client.end_transaction("MyView")
    spans = elasticapm_client.events[constants.SPAN]

    headers = waiting_httpserver.requests[0].headers
    assert constants.TRACEPARENT_HEADER_NAME in headers
    trace_parent = TraceParent.from_string(
        headers[constants.TRACEPARENT_HEADER_NAME], tracestate_string=headers[constants.TRACESTATE_HEADER_NAME]
    )
    assert trace_parent.trace_id == transaction.trace_parent.trace_id
    if sampled:
        assert trace_parent.span_id == spans[0]["id"]
    else:
        assert trace_parent.tracestate_dict[constants.TRACESTATE.SAMPLE_RATE] == "0"
        assert trace_parent.span_id == transaction.id


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        pytest.param({"use_elastic_traceparent_header": True}, id="use_elastic_traceparent_header-True"),
        pytest.param({"use_elastic_traceparent_header": False}, id="use_elastic_traceparent_header-False"),
    ],
    indirect=True,
)
async def test_trace_parent_propagation_unsampled(instrument, event_loop, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    transaction_object = elasticapm_client.begin_transaction("transaction")
    transaction_object.is_sampled = False
    async with aiohttp.ClientSession() as session:
        async with session.get(waiting_httpserver.url) as resp:
            status = resp.status
            text = await resp.text()
    elasticapm_client.end_transaction("MyView")
    spans = elasticapm_client.events[constants.SPAN]

    assert not spans

    headers = waiting_httpserver.requests[0].headers
    assert constants.TRACEPARENT_HEADER_NAME in headers
    trace_parent = TraceParent.from_string(
        headers[constants.TRACEPARENT_HEADER_NAME], tracestate_string=headers[constants.TRACESTATE_HEADER_NAME]
    )
    assert trace_parent.trace_id == transaction_object.trace_parent.trace_id
    assert trace_parent.span_id == transaction_object.id
    assert not trace_parent.trace_options.recorded
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict

    if elasticapm_client.config.use_elastic_traceparent_header:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME in headers
        assert headers[constants.TRACEPARENT_HEADER_NAME] == headers[constants.TRACEPARENT_LEGACY_HEADER_NAME]
    else:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME not in headers


@pytest.mark.parametrize(
    "is_sampled", [pytest.param(True, id="is_sampled-True"), pytest.param(False, id="is_sampled-False")]
)
async def test_tracestate_propagation(instrument, event_loop, elasticapm_client, waiting_httpserver, is_sampled):
    traceparent = TraceParent.from_string(
        "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03", "foo=bar,baz=bazzinga"
    )

    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    transaction_object = elasticapm_client.begin_transaction("transaction", trace_parent=traceparent)
    transaction_object.is_sampled = is_sampled
    async with aiohttp.ClientSession() as session:
        async with session.get(waiting_httpserver.url) as resp:
            status = resp.status
            text = await resp.text()
    elasticapm_client.end_transaction("MyView")
    headers = waiting_httpserver.requests[0].headers
    assert headers[constants.TRACESTATE_HEADER_NAME] == "foo=bar,baz=bazzinga"
